package combinatorParsing

/** A mixin for GenericParsers of parsers that build some sort of output.
 */
trait BuildingParsers extends GenericParsers {

  /** The parser's output is unspecified, needs to be given in subclasses */
  type Output

  /** A class for non-terminals; these consist of an output value and a position */
  case class NonTerminal(value: Output)(pos: Position) extends Token(pos) {
    override def toString() = value.toString
  }

  /** A hook to generate an output value from a list of tokens.
   *  @param  tokens   A list of tokens (terminals or non-terminals)
   *                   which constitute the right hand side of a production.
   *  @param  pos      The position of the first token in `elems', or, if `elems'
   *                   is empty, the token immediately following it.
   */
  def build(elems: List[Token], pos: Position): Output

  /** A parser generator delimiting whole productions 
   *  This changes the behavior in GenericParsers in that it invokes `build'
   *  to construct an output value.
   *
   *  production(p) succeeds if p succeeds. In this case, production(p) will
   *  generate an output value from the list of tokens produced by `p'.
   */
  override def production(p: Parser) = new Parser {
    def apply(in: Input) = p(in) match {
      case s @ Success(out, in1) => 
        Success(List(NonTerminal(build(out, in.head.pos))(in.head.pos)), in1)
      case failure => failure
    }
  }

  /** Two implicit conversions which are useful for builds
   *  Enable them by adding a line
   * 
   *  import implicitConversions._
   *
   *  in front of the `build' method
   */
  protected object implicitConversions {
    implicit def token2output(n: Token): Output = n.asInstanceOf[NonTerminal].value
    implicit def output2token(n: Output): Token = NonTerminal(n)(NoPosition)
  }

  /** Extract an output from a parse result */
  def output(r: ParseResult): Option[Output] = r match {
    case Success(List(NonTerminal(out)), _) => Some(out)
    case _ => None
  }
}
  
