package combinatorParsing

import CharArrayReader.NewLineCh

/** A class for source positions
 *  @param line   The line number referred to by the position
 *  @param line   The column number referred to by the position
 *  @param source The contents of the source file referred to by the position
 */
case class Position(line: int, column: int, source: Array[char]) {

  val EOL = System.getProperty("line.separator", "\n")

  /** <p>
   *    Returns short form of position as a string:
   *    <code>&lt;line&gt;.&lt;column&gt;</code>.
   *  </p>
   */
  override def toString = "" + line + "." + column

  /** <p>
   *   Returns long form of position as a string: 
   *   The contents of the source line referred to it, with an ^ symbol
   *   in the following line pointing out the column
   *  </p>
   *  <p>
   *    Example:
   *  </p>
   *  <pre>
   *    List(<b>this</b>, is, an, example, line)
   *             ^</pre>
   */
  def longString = 
    lineContents(line) + EOL + indent(column) + "^"

  private def indent(col: int) = {
    val chars = new StringBuffer
    for (val i <- 1 until col) chars append ' '
    chars.toString
  }

  def <(that: Position) = {
    assert(this.source eq that.source)
    this.line < that.line || 
    this.line == that.line && this.column < that.column
  }

  private def lineContents(lnum: int) = {
    var i = 0
    var l = 1
    while (i < source.length && l < lnum) {
      while (i < source.length && source(i) != NewLineCh) i = i + 1
      i = i + 1
      l = l + 1
    }
    var chars = new StringBuffer
    while (i < source.length && source(i) != NewLineCh) {
      chars append source(i)
      i = i + 1
    }
    chars.toString
  }
}

object NoPosition extends Position(0, 0, null)
    
