package tapl5

import combinatorParsing.{BuildingParsers, Position, TokenizingParsers}

/** This object implements a parser and evaluator for the
 *  simply typed lambda calculus found in Chapter 9 of
 *  the TAPL book.
 */
object InferencedSTLC extends TokenizingParsers with BuildingParsers with TestSuite  {
  delimiters.incl('(', ')', '\\', '.', ':', '{', '}', ',', '*', '=')
  keywords.incl("Bool", "Nat", "true", "false", "if", "then", "else", "succ",
                "pred", "iszero", "let", "in", "->")

  /** <pre>
   *  Term     ::= SimpleTerm { SimpleTerm }</pre>
   */
  def Term: Parser = production(
      SimpleTerm & rep(SimpleTerm)
    | failure("illegal start of term"))

  /** <pre>
   *  SimpleTerm ::= "true"
   *               | "false"
   *               | number
   *               | "succ" Term
   *               | "pred" Term
   *               | "iszero" Term
   *               | "if" Term "then" Term "else" Term
   *               | ident
   *               | "\" ident [":" Type] "." Term
   *               | "(" Term ")"
   *               | "let" ident [":" Type] "=" Term "in" Term</pre>
   */
  def SimpleTerm: Parser = production(
      kw("true")
    | kw("false")
    | numericLit
    | kw("succ") & Term
    | kw("pred") & Term
    | kw("iszero") & Term
    | kw("if") & Term & kw("then") & Term & kw("else") & Term
    | ident
    | kw("\\") & ident & opt(kw(":") & Type) & kw(".") & Term
    | kw("(") & Term & kw(")")
    | kw("let") & ident & kw("=") & Term & kw("in") & Term
    | failure("illegal start of simple term"))

  /** <pre>
   *  Type       ::= SimpleType { "->" Type }</pre>
   */
  def Type: Parser = production(
      SimpleType & rep(kw("->") & Type)
    | failure("illegal start of type"))

  /** <pre>
   *  SimpleType ::= BaseType [ "*" Type ]</pre>
   */
  def SimpleType: Parser = production(
      BaseType & opt(kw("*") & Type)
    | failure("illegal start of simple type"))

  /** <pre>
   *  BaseType ::= "Bool" | "Nat" | "(" Type ")"</pre>
   */
  def BaseType: Parser = production(
      kw("Bool") | kw("Nat") | kw("(") & Type & kw(")"))

  type Output = Term

  import implicitConversions._

  override def build(xs: List[Token], pos: Position): Output = {
    val t: Term = xs match {
      case KW("true") :: Nil =>
        True
      case KW("false") :: Nil =>
        False
      case NumericLit(chars) :: Nil =>
        def lit2Num(n: Int): Term = if (n == 0) Zero else Succ(lit2Num(n - 1))
        lit2Num(chars.toInt)
      case KW("succ") :: t :: Nil =>
        Succ(t)
      case KW("pred") :: t :: Nil =>
        Pred(t)
      case KW("iszero") :: t :: Nil =>
        IsZero(t)
      case KW("if") :: t1 :: KW("then") :: t2 :: KW("else") :: t3 :: Nil =>
        If(t1, t2, t3)
      // lambda terms
      case Identifier(ident) :: Nil =>
        Var(ident)
      case KW("\\") :: Identifier(ident) :: KW(":") :: NonTerminal(tp: TypeTree) ::
           KW(".") :: t :: Nil =>
        Abs(ident, tp, t)
      case KW("\\") :: Identifier(ident) :: KW(".") :: t :: Nil =>
        Abs(ident, EmptyType, t)
      case KW("(") :: t :: KW(")") :: Nil =>
        t
      case KW("let") :: Identifier(ident) ::
           KW("=") :: t1 :: KW("in") :: t2 :: Nil =>
        Let(ident, t1, t2)
      case KW("Bool") :: Nil =>
        BoolType
      case KW("Nat") :: Nil =>
        NatType
      case NonTerminal(tp1: TypeTree) :: KW("->") :: NonTerminal(tp2: TypeTree) :: Nil =>
        FunType(tp1, tp2)
      case NonTerminal(t1: TypeTree) :: Nil => t1
      case ts =>
        ts.reduceLeft[Token]({ (t1, t2) => App(t1, t2) })
    }
    t.setPos(pos)
  }

  def parse(in: String): Option[Output] = {
    val res = run(Term, in.replaceAll("->", " -> "))
    res match {
      case Success(_, _) => output(res)
      case Failure(_, _) => Console.println(res); None
    }
  }

//  object Inferencer extends TypeInferencers {}
  val Inferencer = new TwoPhaseInferencer

  def showType(s: String) = parse(s) match {
    case Some(t) => Inferencer.showType(t)
    case None    => ()
  }

  def main(args: Array[String]) {
    if (args.length < 1) {
      Console.println("usage: tapl5.InferencedSTLC <file>")
      exit(1)
    }
    check(args(0))
  }
}
